//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Definition and implementation of a "range finder": a template class
// that remembers the N lowest and the N highest values of an abstract
// data type T
// -------------------------------------------------------------------
// $Id: sides.hh,v 1.1 2003/09/23 11:21:25 loreti Exp $
// -------------------------------------------------------------------

#ifndef SIDES_HH_
#define SIDES_HH_

#include <set>
#include <stdexcept>

// The "range finder" is implemented as a template class; the template
// arguments are the abstract data type to be handled, T, and the
// number of elements to be remembered, N (defaults to one).

template <typename T, unsigned int N = 1> class Sides;

// Here follows a specialization of the class for N = 1 (only a single
// value has to be remembered).

template <typename T> class Sides<T,1> {
private:
  T    _minVal;                 // Minimum value found so far
  T    _maxVal;                 // Maximum value found so far
  bool _noVal;                  // Have we seen a value yet?

public:
  Sides() : _noVal(true) {}     // Ctor
  ~Sides() {}                   // Dtor

  inline void enter(const T &); // To handle a new value
  inline const T & getMin();    // Accessor to the minimum value found
  inline const T & getMax();    // Accessor to the maximum value found
};

// Implementation of the above methods for N = 1

template <typename T> inline void Sides<T,1>::enter(
  const T & value
) {
  if (_noVal) {
    _noVal  = false;
    _minVal = _maxVal = value;
  } else {
    if (value < _minVal) _minVal = value;
    if (value > _maxVal) _maxVal = value;
  }
}

template <typename T> inline const T & Sides<T,1>::getMin() {
  if (_noVal) throw std::runtime_error("No values known");
  return _minVal;
}

template <typename T> inline const T & Sides<T,1>::getMax() {
  if (_noVal) throw std::runtime_error("No values known");
  return _maxVal;
}

// The general implementation; the lowest and the highest values are
// saved in a set<T>; a multiset is needed, since some values may be
// repeated.

template <typename T, unsigned int N> class Sides {
private:
  std::multiset<T> _minValues;  // The lowest values
  std::multiset<T> _maxValues;  // The highest values
  unsigned int     _nValues;    // No. of values seen so far

public:
  Sides() : _nValues(0) {}      // Ctor
  ~Sides() {}                   // Dtor

  inline void enter(const T &);
  inline const T & getMin();
  inline const T & getMax();

  // Two additional methods to retrieve the whole sets of the lowest
  // and the highest values

  inline const std::multiset<T> * getLo();
  inline const std::multiset<T> * getHi();
};

template <typename T, unsigned int N> inline void Sides<T,N>::enter(
  const T & value
) {
  if (_nValues < N) {
    ++_nValues;
    _minValues.insert(value);
    _maxValues.insert(value);
  } else {
    if (value <= *(_minValues.begin())) {
      _minValues.erase( --_minValues.end() );
      _minValues.insert(value);
    }
    if (value >= *(_maxValues.rbegin())) {
      _maxValues.erase( _maxValues.begin() );
      _maxValues.insert(value);
    }
  }
}

template <typename T, unsigned int N> inline const T & Sides<T,N>::getMin() {
  if (_nValues == 0) throw std::runtime_error("No values known");
  return *(_minValues.begin());
}

template <typename T, unsigned int N> inline const T & Sides<T,N>::getMax() {
  if (_nValues == 0) throw std::runtime_error("No values known");
  return *(_maxValues.rbegin());
}

template <typename T, unsigned int N>
inline const std::multiset<T> * Sides<T,N>::getLo() {
  if (_nValues == 0) throw std::runtime_error("No values known");
  return &_minValues;
}

template <typename T, unsigned int N>
inline const std::multiset<T> * Sides<T,N>::getHi() {
  if (_nValues == 0) throw std::runtime_error("No values known");
  return &_maxValues;
}

#endif // SIDES_HH_
